import 'dart:ffi' as ffi;

import 'package:ffi/ffi.dart' as ffi;
import 'package:more/more.dart' show Tuple2;

import 'bindings.dart' as ffi;
import 'dylib.dart';
import 'int8str.dart';
import 'settings_schema.dart';
import 'variant.dart';
import 'variant_type.dart';

export 'package:more/more.dart' show Tuple2;

class GSettings {
  final ffi.Pointer<ffi.GSettings> _ptr;

  factory GSettings({required String schemaId, String? path}) {
    if (path == null) return GSettings._new(schemaId);
    return GSettings._newWithPath(schemaId, path);
  }

  GSettings.fromPointer(this._ptr);
  ffi.Pointer<ffi.GSettings> toPointer() => _ptr;

  factory GSettings._new(String schema) {
    final cschema = schema.newInt8String();
    final settings = dylib.g_settings_new(cschema);
    cschema.dispose();
    return GSettings.fromPointer(settings);
  }

  factory GSettings._newWithPath(String schema, String path) {
    final cschema = schema.newInt8String();
    final cpath = path.newInt8String();
    final gsettings = dylib.g_settings_new_with_path(cschema, cpath);
    cschema.dispose();
    cpath.dispose();
    return GSettings.fromPointer(gsettings);
  }

  void sync() => dylib.g_settings_sync();

  void dispose() => dylib.g_object_unref(_ptr.cast());

  String get schemaId => _getProperty('schema-id');
  String get path => _getProperty('path');

  String _getProperty(String name) {
    final value = ffi.calloc.allocate<ffi.GValue>(1);
    dylib.g_value_init(value, ffi.G_TYPE_STRING);
    final property = name.newInt8String();
    dylib.g_object_get_property(_ptr.cast(), property, value);
    final str = dylib.g_value_get_string(value).toDartString();
    property.dispose();
    ffi.calloc.free(value);
    return str;
  }

  Object? value(String key) {
    final ckey = key.newInt8String();
    final ffi.Pointer<ffi.GVariant> variant =
        dylib.g_settings_get_value(_ptr, ckey);
    ckey.dispose();
    return variant.toObject();
  }

  String _typeName(String key) {
    final schema = GSettingsSchema.fromId(schemaId);
    final schemaKey = schema.key(key);
    final typeName = schemaKey.typeName;
    schemaKey.dispose();
    return typeName;
  }

  bool setValue(String key, Object value) {
    final type = GVariantType.fromTypeName(_typeName(key));
    final cvar = value.toVariant(type).toPointer();
    final ckey = key.newInt8String();
    final res = dylib.g_settings_set_value(_ptr, ckey, cvar);
    type.dispose();
    ckey.dispose();
    return res != 0;
  }

  bool isWritable(String key) {
    final ckey = key.newInt8String();
    final res = dylib.g_settings_is_writable(_ptr, ckey);
    ckey.dispose();
    return res != 0;
  }

  void delay() {
    dylib.g_settings_delay(_ptr);
  }

  void apply() {
    dylib.g_settings_apply(_ptr);
  }

  void revert() {
    dylib.g_settings_revert(_ptr);
  }

  bool hasUnapplied() {
    return dylib.g_settings_get_has_unapplied(_ptr) != 0;
  }

  GSettings child(String name) {
    final cname = name.newInt8String();
    final gsettings = dylib.g_settings_get_child(_ptr, cname);
    cname.dispose();
    return GSettings.fromPointer(gsettings);
  }

  void reset(String key) {
    final ckey = key.newInt8String();
    dylib.g_settings_reset(_ptr, ckey);
    ckey.dispose();
  }

  Object? defaultValue(String key) {
    final ckey = key.newInt8String();
    final cvariant = dylib.g_settings_get_default_value(_ptr, ckey);
    final value = cvariant.toObject();
    ckey.dispose();
    return value;
  }

  List<String> children() {
    final list = <String>[];
    final cchildren = dylib.g_settings_list_children(_ptr);
    var i = 0;
    while (cchildren[i] != ffi.nullptr) {
      list.add(cchildren[i].toDartString());
      ++i;
    }
    dylib.g_strfreev(cchildren);
    return list;
  }

  bool? boolValue(String key) => value(key) as bool?;
  int? intValue(String key) => value(key) as int?;
  double? doubleValue(String key) => value(key) as double?;
  String? stringValue(String key) => value(key) as String?;
  Object? maybeValue(String key) => value(key);
  Map<Object, Object?>? dictValue(String key) =>
      value(key) as Map<Object, Object?>?;
  List<Object?>? arrayValue(String key) => value(key) as List<Object?>?;
  List<int?>? byteArrayValue(String key) => arrayValue(key)?.cast();
  List<String?>? stringArrayValue(String key) => arrayValue(key)?.cast();
  Tuple2<Object?, Object?>? tupleValue(String key) =>
      value(key) as Tuple2<Object?, Object?>?;
}
