import 'dart:ffi' as ffi;

import 'package:ffi/ffi.dart' as ffi;
import 'package:more/more.dart';

import 'bindings.dart' as ffi;
import 'dylib.dart';
import 'int8str.dart';
import 'variant_builder.dart';
import 'variant_type.dart';

class GVariant {
  final ffi.Pointer<ffi.GVariant> _ptr;

  GVariant.fromPointer(ffi.Pointer<ffi.GVariant> variant)
      : assert(variant != ffi.nullptr),
        _ptr = variant;

  ffi.Pointer<ffi.GVariant> toPointer() => _ptr;

  // ignore: avoid_positional_boolean_parameters
  GVariant.fromBoolean(bool value)
      : _ptr = dylib.g_variant_new_boolean(value ? 1 : 0);
  GVariant.fromByte(int value) : _ptr = dylib.g_variant_new_byte(value);
  GVariant.fromInt16(int value) : _ptr = dylib.g_variant_new_int16(value);
  GVariant.fromUInt16(int value) : _ptr = dylib.g_variant_new_uint16(value);
  GVariant.fromInt32(int value) : _ptr = dylib.g_variant_new_int32(value);
  GVariant.fromUInt32(int value) : _ptr = dylib.g_variant_new_uint32(value);
  GVariant.fromInt64(int value) : _ptr = dylib.g_variant_new_int64(value);
  GVariant.fromUInt64(int value) : _ptr = dylib.g_variant_new_uint64(value);
  GVariant.fromDouble(double value) : _ptr = dylib.g_variant_new_double(value);
  GVariant.fromString(String value)
      : _ptr = dylib.g_variant_new_take_string(value.newInt8String());

  GVariantType get type => _ptr.type;
  String get typeName => _ptr.typeName;

  bool toBoolean() => _ptr.toBoolean();
  int toByte() => _ptr.toByte();
  int toInt16() => _ptr.toInt16();
  int toUInt16() => _ptr.toUInt16();
  int toInt32() => _ptr.toInt32();
  int toUInt32() => _ptr.toUInt64();
  int toInt64() => _ptr.toInt64();
  int toUInt64() => _ptr.toUInt64();
  double toDouble() => _ptr.toDouble();
  String toDartString() => _ptr.toString();
  List<Object?> toList() => _ptr.toList();
  Map<Object, Object?> toMap() => _ptr.toMap();
  Tuple2<Object?, Object?> toTuple() => _ptr.toTuple();

  Object? toObject() => _ptr.toObject();

  void dispose() => _ptr.dispose();
}

extension GVariantPointer on ffi.Pointer<ffi.GVariant> {
  GVariantType get type =>
      GVariantType.fromPointer(dylib.g_variant_get_type(this));
  String get typeName => dylib.g_variant_get_type_string(this).toDartString();

  bool toBoolean() => dylib.g_variant_get_boolean(this) != 0;
  int toByte() => dylib.g_variant_get_byte(this);
  int toInt16() => dylib.g_variant_get_int16(this);
  int toUInt16() => dylib.g_variant_get_uint16(this);
  int toInt32() => dylib.g_variant_get_int32(this);
  int toUInt32() => dylib.g_variant_get_uint32(this);
  int toInt64() => dylib.g_variant_get_int64(this);
  int toUInt64() => dylib.g_variant_get_uint64(this);
  double toDouble() => dylib.g_variant_get_double(this);

  String toDartString() {
    final ffi.Pointer<ffi.Int32> size = ffi.calloc<ffi.Int32>();
    final ffi.Pointer<ffi.Int8> cstr = dylib.g_variant_get_string(this, size);
    final String string = cstr.toDartString(length: size.value);
    ffi.calloc.free(size);
    return string;
  }

  List<Object?> toList() {
    return List<Object?>.generate(_childCount, (int index) {
      return _childObject(index);
    });
  }

  MapEntry<Object, Object?> toMapEntry() {
    final key = _childObject(0);
    final value = _childObject(1);
    return MapEntry(key!, value);
  }

  Map<Object, Object?> toMap() {
    final map = <Object, Object?>{};
    for (var i = 0; i < _childCount; ++i) {
      final entry = _childObject(i) as MapEntry<Object, Object?>?;
      map[entry!.key] = entry.value;
    }
    return map;
  }

  Tuple2<Object?, Object?> toTuple() {
    return Tuple2(_childObject(0), _childObject(1));
  }

  Object? toMaybe() {
    final ffi.Pointer<ffi.GVariant> variant = dylib.g_variant_get_maybe(this);
    if (variant == ffi.nullptr) return null;
    final maybe = variant.toObject();
    variant.dispose();
    return maybe;
  }

  Object? toObject() {
    switch (type.typeName) {
      case GVariantTypeName.boolean:
        return toBoolean();
      case GVariantTypeName.byte:
        return toByte();
      case GVariantTypeName.int16:
        return toInt16();
      case GVariantTypeName.uint16:
        return toUInt16();
      case GVariantTypeName.int32:
        return toInt32();
      case GVariantTypeName.uint32:
        return toUInt32();
      case GVariantTypeName.int64:
        return toInt64();
      case GVariantTypeName.uint64:
        return toUInt64();
      case GVariantTypeName.double:
        return toDouble();
      case GVariantTypeName.string:
        return toDartString();
      default:
        if (type.isMaybe) return toMaybe();
        if (type.isArray) {
          final elementType = GVariantType.fromTypeName(type.elementType);
          final isDictEntry = elementType.isDictEntry;
          elementType.dispose();
          if (isDictEntry) return toMap();
          return toList();
        }
        if (type.isDictEntry) return toMapEntry();
        if (type.isTuple) return toTuple();
        return null;
    }
  }

  int get _childCount => dylib.g_variant_n_children(this);

  Object? _childObject(int index) {
    final ffi.Pointer<ffi.GVariant> variant =
        dylib.g_variant_get_child_value(this, index);
    final child = variant.toObject();
    variant.dispose();
    return child;
  }

  void dispose() => dylib.g_variant_unref(this);
}

extension GVariantObject on Object {
  GVariant toVariant(GVariantType type) {
    switch (type.typeName) {
      case GVariantTypeName.boolean:
        return GVariant.fromBoolean(this as bool);
      case GVariantTypeName.byte:
        return GVariant.fromByte(this as int);
      case GVariantTypeName.int16:
        return GVariant.fromInt16(this as int);
      case GVariantTypeName.uint16:
        return GVariant.fromUInt16(this as int);
      case GVariantTypeName.int32:
        return GVariant.fromInt32(this as int);
      case GVariantTypeName.uint32:
        return GVariant.fromUInt32(this as int);
      case GVariantTypeName.int64:
        return GVariant.fromInt64(this as int);
      case GVariantTypeName.uint64:
        return GVariant.fromUInt64(this as int);
      case GVariantTypeName.double:
        return GVariant.fromDouble(this as double);
      case GVariantTypeName.string:
        return GVariant.fromString(this as String);
      default:
        if (type.isMaybe) {
          return toVariantMaybe(type);
        }
        if (type.isDict) {
          return (this as Map<Object, Object>).toVariantDict(type);
        }
        if (type.isArray) {
          return (this as List<Object>).toVariantArray(type);
        }
        if (type.isDictEntry) {
          return (this as MapEntry<Object, Object>).toVariantDictEntry(type);
        }
        if (type.isTuple) {
          return (this as Tuple2<Object, Object>).toVariantTuple(type);
        }
        throw UnimplementedError('Unknown type: ${type.typeName}');
    }
  }

  GVariant toVariantMaybe(GVariantType type) {
    final builder = GVariantBuilder.fromType(type);
    final elementType = GVariantType.fromTypeName(type.elementType);
    builder.addValue(toVariant(elementType));
    final maybe = builder.end();
    builder.dispose();
    elementType.dispose();
    return maybe;
  }
}

extension GVariantList on List<Object> {
  GVariant toVariantArray(GVariantType type) {
    final builder = GVariantBuilder.fromType(type);
    final elementType = GVariantType.fromTypeName(type.elementType);
    for (final item in this) {
      builder.addValue(item.toVariant(elementType));
    }
    final variant = builder.end();
    builder.dispose();
    elementType.dispose();
    return variant;
  }
}

extension GVariantDict on Map<Object, Object> {
  GVariant toVariantDict(GVariantType type) {
    final builder = GVariantBuilder.fromType(type);
    final elementType = GVariantType.fromTypeName(type.elementType);
    GVariantType? keyType, valueType;
    for (final entry in entries) {
      keyType ??= GVariantType.fromTypeName(elementType.keyType);
      valueType ??= GVariantType.fromTypeName(elementType.valueType);
      builder.addValue(entry.toVariantDictEntry(elementType));
    }
    final dict = builder.end();
    builder.dispose();
    keyType?.dispose();
    valueType?.dispose();
    elementType.dispose();
    return dict;
  }
}

extension GVariantDictEntry on MapEntry<Object, Object> {
  GVariant toVariantDictEntry(GVariantType type) {
    final builder = GVariantBuilder.fromType(type);
    final keyType = GVariantType.fromTypeName(type.keyType);
    final valueType = GVariantType.fromTypeName(type.valueType);
    builder.addValue(key.toVariant(keyType));
    builder.addValue(value.toVariant(valueType));
    final entry = builder.end();
    builder.dispose();
    keyType.dispose();
    valueType.dispose();
    return entry;
  }
}

extension GVariantTuple on Tuple2<Object, Object> {
  GVariant toVariantTuple(GVariantType type) {
    assert(type.itemTypes.length == 2);
    final builder = GVariantBuilder.fromType(type);
    final firstType = GVariantType.fromTypeName(type.itemTypes.first);
    final secondType = GVariantType.fromTypeName(type.itemTypes.last);
    builder.addValue(first.toVariant(firstType));
    builder.addValue(second.toVariant(secondType));
    final tuple = builder.end();
    builder.dispose();
    firstType.dispose();
    secondType.dispose();
    return tuple;
  }
}
