import 'dart:ffi' as ffi;

import 'bindings.dart' as ffi;
import 'dylib.dart';
import 'int8str.dart';

class GVariantType {
  final ffi.Pointer<ffi.Int8> _cname;
  ffi.Pointer<ffi.GVariantType>? __ptr;

  GVariantType.fromTypeName(String typeName)
      : assert(GVariantTypeName.isValid(typeName)),
        _cname = typeName.newInt8String();

  GVariantType.fromPointer(ffi.Pointer<ffi.GVariantType> ptr)
      : assert(ptr != ffi.nullptr),
        _cname = ptr.typeName.newInt8String(),
        __ptr = ptr;

  ffi.Pointer<ffi.GVariantType> toPointer() => _ptr;

  void dispose() {
    _cname.dispose();
    __ptr?.dispose();
  }

  ffi.Pointer<ffi.GVariantType> get _ptr {
    return __ptr ??= dylib.g_variant_type_new(_cname);
  }

  String get typeName => _cname.toDartString();

  bool get isArray => _ptr.isArray;
  bool get isBasic => _ptr.isBasic;
  bool get isContainer => _ptr.isContainer;
  bool get isDefinite => _ptr.isDefinive;
  bool get isDictEntry => _ptr.isDictEntry;
  bool get isMaybe => _ptr.isMaybe;
  bool get isTuple => _ptr.isTuple;
  bool get isVariant => _ptr.isVariant;

  bool get isDict {
    if (!isArray) return false;
    final etype = GVariantType.fromTypeName(elementType);
    final isDictEntry = etype.isDictEntry;
    etype.dispose();
    return isDictEntry;
  }

  // array, maybe
  String get elementType => _ptr.elementType;

  // dict entry
  String get keyType => _ptr.keyType;
  String get valueType => _ptr.valueType;

  // tuple
  List<String> get itemTypes => _ptr.itemTypes;
}

extension GVariantTypePointer on ffi.Pointer<ffi.GVariantType> {
  String get typeName {
    final length = dylib.g_variant_type_get_string_length(this);
    final ctype = dylib.g_variant_type_peek_string(this);
    return ctype.toDartString(length: length);
  }

  void dispose() => dylib.g_variant_type_free(this);

  bool get isArray => dylib.g_variant_type_is_array(this) != 0;
  bool get isBasic => dylib.g_variant_type_is_basic(this) != 0;
  bool get isContainer => dylib.g_variant_type_is_container(this) != 0;
  bool get isDefinive => dylib.g_variant_type_is_definite(this) != 0;
  bool get isDictEntry => dylib.g_variant_type_is_dict_entry(this) != 0;
  bool get isMaybe => dylib.g_variant_type_is_maybe(this) != 0;
  bool get isTuple => dylib.g_variant_type_is_tuple(this) != 0;
  bool get isVariant => dylib.g_variant_type_is_variant(this) != 0;

  // array, maybe
  String get elementType {
    assert(isArray || isMaybe);
    return dylib.g_variant_type_element(this).typeName;
  }

  // dict entry
  String get keyType {
    assert(isDictEntry);
    return dylib.g_variant_type_key(this).typeName;
  }

  // dict entry
  String get valueType {
    assert(isDictEntry);
    return dylib.g_variant_type_value(this).typeName;
  }

  // tuple
  List<String> get itemTypes {
    assert(isTuple);
    final types = <String>[];
    var next = dylib.g_variant_type_first(this);
    while (next != ffi.nullptr) {
      types.add(next.typeName);
      next = dylib.g_variant_type_next(next);
    }
    return types;
  }
}

abstract class GVariantTypeName {
  const GVariantTypeName._();

  static bool isValid(String typeName) {
    final cname = typeName.newInt8String();
    final res = dylib.g_variant_type_string_is_valid(cname);
    cname.dispose();
    return res != 0;
  }

  static const String boolean = 'b';
  static const String byte = 'y';
  static const String int16 = 'n';
  static const String uint16 = 'q';
  static const String int32 = 'i';
  static const String uint32 = 'u';
  static const String int64 = 'x';
  static const String uint64 = 't';
  static const String double = 'd';
  static const String string = 's';
  static const String objectPath = 'o';
  static const String signature = 'g';
  static const String variant = 'v';
  static const String handle = 'h';
  static const String unit = '()';
  static const String any = '*';
  static const String basic = '?';
  static const String maybe = 'm*';
  static const String array = 'a*';
  static const String tuple = 'r';
  static const String dictEntry = '{?*}';
  static const String dictionary = 'a{?*}';
  static const String stringArray = 'as';
  static const String objectPathArray = 'ao';
  static const String byteString = 'ay';
  static const String byteStringArray = 'aay';
  static const String varDict = 'a{sv}';
}
