import 'dart:ffi';
import 'dart:io';

import 'package:gsettings/gsettings.dart';
import 'package:gsettings/src/int8str.dart';
import 'package:gsettings/src/dylib.dart';

import 'package:more/more.dart';
import 'package:path/path.dart' as path;
import 'package:test/test.dart';

void setenv(String key, String value) {
  final ckey = key.newInt8String();
  final cvalue = value.newInt8String();
  dylib.g_setenv(ckey, cvalue, 1);
  ckey.dispose();
  cvalue.dispose();
}

String testSchemaPath() {
  var dir = Directory.current;
  // https://github.com/flutter/flutter/issues/20907
  if (dir.path.endsWith('/test')) {
    dir = Directory.current.parent;
  }
  return path.join(dir.path, 'test/schemas/');
}

void main() {
  late GSettings settings;

  setUp(() {
    setenv('GSETTINGS_SCHEMA_DIR', testSchemaPath());
    settings = GSettings(schemaId: 'gsettings.dart.test');
  });

  tearDown(() {
    settings.dispose();
  });

  test('bool', () {
    const key = 'bool-value';

    expect(settings.defaultValue(key), isTrue);
    expect(settings.setValue(key, false), isTrue);
    expect(settings.boolValue(key), isFalse);
  });

  test('byte', () {
    const key = 'byte-value';

    expect(settings.defaultValue(key), equals(255));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('int16', () {
    const key = 'int16-value';

    expect(settings.defaultValue(key), equals(32767));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('uint16', () {
    const key = 'uint16-value';

    expect(settings.defaultValue(key), equals(65535));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('int32', () {
    const key = 'int32-value';

    expect(settings.defaultValue(key), equals(2147483647));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('uint32', () {
    const key = 'uint32-value';

    // ### TODO: expect(settings.defaultValue(key), equals(4294967295));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('int64', () {
    const key = 'int64-value';

    // ### TODO: expect(settings.defaultValue(key), equals(9223372036854775807));
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('double', () {
    const key = 'double-value';

    expect(settings.defaultValue(key), equals(123456.789));
    expect(settings.setValue(key, 123.456), isTrue);
    expect(settings.doubleValue(key), equals(123.456));
  });

  test('string', () {
    const key = 'string-value';

    expect(settings.defaultValue(key), equals('foo'));
    expect(settings.setValue(key, 'bar'), isTrue);
    expect(settings.stringValue(key), equals('bar'));
  });

  test('int-array', () {
    const key = 'array-int-values';
    const defaults = <int>[1, 2, 3];
    const ints = <int>[4, 5, 6];

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, ints), isTrue);
    expect(settings.arrayValue(key), equals(ints));
    expect(settings.byteArrayValue(key), equals(ints));
  });

  test('string-array', () {
    const key = 'array-string-values';
    const defaults = <String>['f', 'o', 'o'];
    const strings = <String>['foo', 'bar', 'baz'];

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, strings), isTrue);
    expect(settings.arrayValue(key), equals(strings));
    expect(settings.stringArrayValue(key), equals(strings));
  });

  test('int-int-dict', () {
    const key = 'dict-int-int-values';
    const defaults = <int, int>{1: 4, 2: 5, 3: 6};
    const entries = <int, int>{1: 7, 2: 8, 3: 9};

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, entries), isTrue);
    expect(settings.dictValue(key), equals(entries));
  });

  test('string-int-dict', () {
    const key = 'dict-string-int-values';
    const defaults = <String, int>{'a': 1, 'b': 2, 'c': 3};
    const entries = <String, int>{'a': 4, 'b': 5, 'c': 6};

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, entries), isTrue);
    expect(settings.dictValue(key), equals(entries));
  });

  test('string-string-dict', () {
    const key = 'dict-string-string-values';
    const defaults = <String, String>{'a': 'd', 'b': 'e', 'c': 'f'};
    const entries = <String, String>{'a': 'g', 'b': 'h', 'c': 'i'};

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, entries), isTrue);
    expect(settings.dictValue(key), equals(entries));
  });

  test('string-int-tuple', () {
    const key = 'tuple-string-int-values';
    const defaults = Tuple2<String, int>('a', 1);
    const entries = Tuple2<String, int>('b', 2);

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, entries), isTrue);
    expect(settings.tupleValue(key), equals(entries));
  });

  test('string-string-tuple', () {
    const key = 'tuple-string-string-values';
    const defaults = Tuple2<String, String>('foo', 'bar');
    const entries = Tuple2<String, String>('baz', 'qux');

    expect(settings.defaultValue(key), equals(defaults));
    expect(settings.setValue(key, entries), isTrue);
    expect(settings.tupleValue(key), equals(entries));
  });

  test('nullable-int', () {
    const key = 'nullable-int';

    expect(settings.defaultValue(key), isNull);
    expect(settings.setValue(key, 123), isTrue);
    expect(settings.intValue(key), equals(123));
  });

  test('nullable-string', () {
    const key = 'nullable-string';

    expect(settings.defaultValue(key), isNull);
    expect(settings.setValue(key, 'foo'), isTrue);
    expect(settings.stringValue(key), equals('foo'));
  });
}
